#@(#) $Header: /home/softools/cvs/softools/code/common/cked/src/ss.tcl,v 1.78 2004/06/22 18:22:18 mark Exp $
#
# This file contains bindings for the spreadsheet widget.
#

foreach binding [bind Ss] {
  bind Ss $binding {}
}
bind Ss <Destroy>          {::ss::destroy %W}
bind Ss <Configure>        {::ss::widget_configured %W %x %y %w %h}

bind Ss <KeyPress>         {::ss::insert %W %A}
bind Ss <BackSpace>        {::ss::backspace %W}
bind Ss <Delete>           {::ss::delete_key %W}
catch {
  bind Ss <KP_Delete>      {::ss::delete_key %W}
}
bind Ss <Escape>           {::ss::escape %W}
bind Ss <<Table_Commit>>   {catch {%W activate active}}

# Ignore all Alt, Meta, and Control keypresses unless explicitly bound.
# Otherwise, if a widget binding for one of these is defined, the
# <KeyPress> class binding will also fire and insert the character,
# which is wrong.  Ditto for Tab.

bind Ss <Alt-KeyPress>     {# nothing}
bind Ss <Meta-KeyPress>    {# nothing}
bind Ss <Control-KeyPress> {# nothing}

# Redefine the "all" binding for <Tab> so that the spreadsheet binding
# takes precedence.

set bind_tab [bind all <Tab>]
bind all <Tab>             "::ss::all_tab %W \"$bind_tab\""

bind Ss <Any-Tab>          {::ss::move_cell %W 1 0}

set bind_backtab [bind all <<PrevWindow>>]
bind all <<PrevWindow>>    "::ss::all_backtab %W \"$bind_backtab\""

bind Ss <<PrevWindow>>     {::ss::move_cell %W -1 0}
bind Ss <Control-Tab>      {::ss::move_cell %W -1 0}

# These bindings differ from the table widget.
# The table widget doesn't support an "edit mode" for these characters.
# Excel supports both a "cell edit mode" and a "table scan mode".
# I think both of these are wrong.  Why not just support cell edit mode?

bind Ss <Up>                 {::ss::up %W}
bind Ss <Shift-Up>           {::ss::shift_up %W}
bind Ss <Down>               {::ss::down %W}
bind Ss <Shift-Down>         {::ss::shift_down %W}
bind Ss <Left>               {::ss::left %W}
bind Ss <Shift-Left>         {::ss::shift_left %W}
bind Ss <Control-Left>       {::ss::control_left %W}
bind Ss <Right>              {::ss::right %W}
bind Ss <Shift-Right>        {::ss::shift_right %W}
bind Ss <Control-Right>      {::ss::control_right %W}
bind Ss <Return>             {::ss::down %W}
bind Ss <Next>               {::ss::page_down %W}
bind Ss <Prior>              {::ss::page_up %W}
bind Ss <Control-Home>       {::ss::control_home %W}
bind Ss <Control-Shift-Home> {::ss::control_shift_home %W}
bind Ss <Control-End>        {::ss::control_end %W}
bind Ss <Control-Shift-End>  {::ss::control_shift_end %W}
bind Ss <1>                  {::ss::b1_pressed_can %W %x %y}
bind Ss <Shift-1>            {;#}
bind Ss <ButtonRelease-1>    {;#}

# We need this binding at this level to make auto-scanning work right.
# (Auto-scanning causes the underlying spreadsheet canvas to be destroyed
# and re-created.  Thus we can't use a tag binding to implement it.)
#
bind Ss <B1-Motion>       {::ss::b1_motion %W %x %y}
bind Ss <Shift-1>         {;#}
bind Ss <Control-1>       {;#}
bind Ss <ButtonRelease-1> {::ss::b1_released %W}
catch {
  bind Ss <KP_Up>        {::ss::up %W}
  bind Ss <KP_Down>      {::ss::down %W}
  bind Ss <KP_Left>      {::ss::left %W}
  bind Ss <KP_Right>     {::ss::right %W}
  bind Ss <KP_Return>    {::ss::down %W}
  bind Ss <KP_Next>      {::ss::page_down %W}
  bind Ss <KP_Prior>     {::ss::page_up %W}
}

# Create the ss namespace, which will be used to create and execute
# bindings on the spreadsheet's underlying canvas
#
namespace eval ss {

# proc:  ::ss::init $w
# title: Initialize internal values of the spreadsheet $w
#
proc init {w} {

  # initialized     -> 1 if these values have been initialized
  # mouse_moved     -> nonzero if the mouse has moved since beginning
  #                    the selection
  # orig_selection  -> the original cell selected in the current selection
  #                    (for the spreadsheet-level selection)
  # prev_selection  -> the cell previously visited during a selection
  #                    (for the spreadsheet-level selection)
  # resize_col      -> the column currently being resized
  # resize_line     -> the canvas ID referencing the visible line used
  #                    to resize a row or column.  "" if not resizing.
  # resize_min      -> the minimum X or Y coordinate that resizing will
  #                    not overlap
  # resize_prev     -> the previous X or Y coordinate where the resize
  #                    was just recorded.  The X coordinate is stored for
  #                    a column resize; the Y coordinate is stored for
  #                    a row resize.  -1 if not resizing.
  # resize_row      -> the row currently being resized
  # select_cell     -> the cell ID that contains the text selection
  # x               -> x coordinate of mouse pointer during a selection
  # y               -> y coordinate of mouse pointer during a selection

  upvar #0 $w v
  array set v {
    initialized        1
    mouse_moved        0
    orig_selection     {}
    prev_selection     {}
    resize_col         {}
    resize_line        {}
    resize_min         -1
    resize_prev        -1
    resize_row         {}
    select_cell        {}
    x                  0
    y                  0
  }

  create_canvas_bindings $w
}

# proc:  ::ss::create_canvas_bindings $w
# title: Create canvas bindings for a particular spreadsheet
#
proc create_canvas_bindings {w} {
  $w canvas bind notresize <1>               {::ss::b1_pressed %W %x %y}
  $w canvas bind notresize <ButtonRelease-1> {::ss::b1_released %W}
  $w canvas bind notresize <Shift-1>         {::ss::shift_b1_pressed %W %x %y}
  $w canvas bind notresize <Control-1>       {::ss::control_b1_pressed %W %x %y}
  $w canvas bind notresize <B1-Leave>        {::ss::b1_left %W %x %y}
  $w canvas bind notresize <B1-Enter>        {::ss::b1_entered %W}

  $w canvas bind vresize <Enter>             {::ss::enter_vresize %W}
  $w canvas bind vresize <Leave>             {::ss::leave_vresize %W}
  $w canvas bind vresize <1>                 {::ss::b1_vresize %W %x %y}
  $w canvas bind vresize <B1-Motion>         {::ss::b1_motion_vresize %W %x %y}
  $w canvas bind vresize <ButtonRelease-1>   {::ss::b1_release_vresize %W}

  $w canvas bind hresize <Enter>             {::ss::enter_hresize %W}
  $w canvas bind hresize <Leave>             {::ss::leave_hresize %W}
  $w canvas bind hresize <1>                 {::ss::b1_hresize %W %x %y}
  $w canvas bind hresize <B1-Motion>         {::ss::b1_motion_hresize %W %x %y}
  $w canvas bind hresize <ButtonRelease-1>   {::ss::b1_release_hresize %W}

  $w canvas bind selrect <Enter>             {::ss::enter_selrect %W}
  $w canvas bind selrect <Leave>             {::ss::leave_selrect %W}
  $w canvas bind selrect <1>                 {::ss::b1_in_selrect %W %x %y}
  $w canvas bind selrect <B1-Motion>         {::ss::b1_motion_selrect %W %x %y}
  $w canvas bind selrect <ButtonRelease-1>   {::ss::b1_release_selrect %W}

  $w canvas bind title <Enter>               {::ss::enter_title %W}
  $w canvas bind title <Leave>               {::ss::leave_title %W}
}

#
# The following procs implement procedures used by the spreadsheet bindings
#

# proc: ::ss::destroy $w
# title: This proc is called when the spreadsheet $w is destroyed
#
proc destroy {w} {
  rename ::$w {}
}

# proc:  ::ss::widget_configured $w $x $y $width $height
# title: Handle a <Configure> binding
#
# This proc is called when the <Configure> binding fires on the SS widget.
# It determines whether it needs to redisplay the widget.
# [SRS3320] The widget is resized to fit its window.
#
proc widget_configured {w x y width height} {
  upvar #0 $w v

  # Initialize the widget-specific state if it hasn't already been initialized

  if {![info exists v(initialized)]} {
    init $w
  }

  set bd [$w canvas cget -bd]
  set hbd [$w canvas cget -highlightthickness]
  set width [expr {$width - 2*$bd - 2*$hbd}]
  set height [expr {$height - 2*$bd - 2*$hbd}]
  if {$width==[$w canvas cget -width] && 
      $height==[$w canvas cget -height]} return
  $w canvas config -width $width -height $height

  # This will cause a designated column to expand or shrink so that the
  # spreadsheet exactly fills the available width.  But this won't happen
  # if the expandable column is also hidden (its width<=0).

  set expand_col [$w cget -expandcol]
  set ewidth [$w width $expand_col]
  set hidden [$w hide get $expand_col]
  if {$ewidth>0 && !$hidden} {
    set ncols [$w cget -cols]
    scan [$w index topleft] {%d,%d} toprow leftcol
    if {$expand_col>=$leftcol && $expand_col<$ncols} {
      set ss_width [$w width 0]
      for {set col $leftcol} {$col<$ncols} {incr col} {
        if {![$w hide get $col]} {
          incr ss_width [$w width $col]
        }
      }
      set diff [expr {$width-$ss_width}]
      $w width $expand_col [expr {$ewidth+$diff}]
    }
  }

  # When reconfiguring, do the display immediately--
  # don't wait until idle time.  This avoids flicker.

  $w redisplay -now -trim
}

# proc:  ::ss::insert $w $string
# title: Insert $string in the active cell at the current position
#
proc insert {w string} {
  if {[$w cget -state]=="disabled"} return

  if {$string!=""} {
    $w insert $string
  }
}

# proc:  ::ss::backspace $w
# title: Handle a backspace
#
# Backspace was pressed.  Delete the character before the insertion
# cursor in the active cell, if there is one.  Or delete the entire
# text selection if it exists.
#
proc backspace {w} {
  if {[$w cget -state]=="disabled"} return

  set idx [$w icursor]
  if {$idx==-1} return

  set sel_list [$w curtextsel]
  set select_cell [lindex $sel_list 0]
  set select_cell_from [lindex $sel_list 1]
  set select_cell_to [lindex $sel_list 2]

  set active_cell [$w index active]
  if {$select_cell!="" && $select_cell==$active_cell && \
      $select_cell_from >= 0 && $select_cell_to >= 0} {
    $w delete active $select_cell_from $select_cell_to
    $w textselection clear
  } elseif {$idx!="" && $idx > 0} {
    $w delete active [expr {$idx - 1}]
  }
}

# proc:  ::ss::delete_key $w
# title: Handle the delete key press
#
# The delete key was pressed.  This key either:
# 
# - deletes the entire table selection, if there is one and it's more than
#   just the active cell; or
# - deletes the intra-cell selection, if there is one; or
# - deletes the key to the right of the cursor, if there is one.
# 
proc delete_key {w} {
  if {[$w cget -state]=="disabled"} return

  set selected [$w curselection]
  set active [$w index active]
  if {$selected!="" && $selected!=$active} {
    $w activate active  ;# commit the active value
    $w selection cleardata
    return
  }

  set sel_list [$w curtextsel]
  if {$sel_list!=""} {
    set select_cell [lindex $sel_list 0]
    set select_cell_from [lindex $sel_list 1]
    set select_cell_to [lindex $sel_list 2]
    $w delete active $select_cell_from $select_cell_to
    $w textselection clear
    return
  }

  set idx [$w icursor]
  if {$idx==-1} return
  $w delete active $idx
}

# proc:  ::ss::escape $w
# title: Handle the escape key press
#
# Escape was pressed.  The active cell's contents will be converted to
# its previous value, before the cell was edited.
#
proc escape {w} {
  if {[$w cget -state]=="disabled"} return

  set cell [$w index active]
  set var  [$w cget -variable]
  if {$var!=""} {
    upvar #0 $var v
    if {[info exists v($cell)]} {
      $w set $cell $v($cell)
    } else {
      $w set $cell {}
    }
  }
}

# proc:  ::ss::all_tab $w $script
# title: Handle all Tab presses
#
# This proc is called whenever Tab is pressed in any window in the application.
# It will check to see whether there is a Ss binding for $w, and if not,
# will execute $script.  This allows the Ss-specific binding to take
# precedence over the "all" binding.
#
proc all_tab {w script} {
  if {[lsearch -exact [bindtags $w] "Ss"]==-1} {
    eval $script
  }
}

# proc:  ::ss::all_backtab $w $script
# title: Handle all Backtab presses
#
# This proc is called whenever Back-Tab is pressed in any window 
# in the application.  It will check to see whether there is a 
# Ss binding for $w, and if not,
# will execute $script.  This allows the Ss-specific binding to take
# precedence over the "all" binding.
#
proc all_backtab {w script} {
  if {[lsearch -exact [bindtags $w] "Ss"]==-1} {
    eval $script
  }
}

# proc:  ::ss::up $w
# title: Handle the Up key
#
# Up moves the active cell one cell higher.
#
proc up {w} {
  if {[$w cget -state]=="disabled"} return
  move_cell $w 0 -1
}

# proc:  ::ss::down $w
# title: Handle the Down key
#
# Down moves the active cell one cell lower.
#
proc down {w} {
  if {[$w cget -state]=="disabled"} return
  move_cell $w 0 1
}

# proc:  ::ss::left $w
# title: Handle the Left key
#
# Left moves the cursor to the left, unless it's at the beginning
# of the text in the active cell; if so, it changes the active cell
# to the left.
#
proc left {w} {
  if {[$w cget -state]=="disabled"} return
  set idx [$w icursor]
  if {$idx <= 0} {
    move_cell $w -1 0
  } else {
    incr idx -1
    $w icursor $idx
  }
}

# proc:  ::ss::right $w
# title: Handle the Right key
#
# Right moves the cursor to the right, unless it's at the end
# of the text in the active cell; if so, it changes the active cell
# to the right.
#
proc right {w} {
  if {[$w cget -state]=="disabled"} return
  set idx [$w icursor]
  if {$idx < 0} {
    move_cell $w 1 0
  } else {
    set end [string length [$w curvalue]]
    if {$idx==$end} {
      move_cell $w 1 0
    } else {
      incr idx
      $w icursor $idx
    }
  }
}

# proc:  ::ss::shift_up $w
# title: Handle the Shift-Up key
#
# Extend the selection upward.
#
proc shift_up {w} {
  if {[$w cget -state]=="disabled"} return
  set bbox [$w selrect bbox]
  if {$bbox==""} return
  foreach {top left bottom right} $bbox break
  set anchor [$w index active]
  if {[scan $anchor {%d,%d} arow acol]!=2} return
  scan [$w index topleft] {%d,%d} vtop vleft
  if {$top<$arow} {
    if {$top==1} return
    incr top -1
    $w selection set $anchor $top,$left
    $w selection set $anchor $top,$right
    $w see $top,$vleft
  } else {
    incr bottom -1
    $w selection set -only $anchor $bottom,$left
    $w selection set $anchor $bottom,$right
    $w see $bottom,$vleft
  }
}

# proc:  ::ss::shift_down $w
# title: Handle the Shift-Down key
#
# Extend the selection downward.
#
proc shift_down {w} {
  if {[$w cget -state]=="disabled"} return
  set bbox [$w selrect bbox]
  if {$bbox==""} return
  foreach {top left bottom right} $bbox break
  set anchor [$w index active]
  if {[scan $anchor {%d,%d} arow acol]!=2} return
  set end [$w index end]
  scan $end {%d,%d} lastrow lastcol
  scan [$w index topleft] {%d,%d} vtop vleft
  if {$bottom>$arow} {
    if {$bottom==$lastrow} return
    incr bottom
    $w selection set $anchor $bottom,$left
    $w selection set $anchor $bottom,$right
    $w see $bottom,$vleft
  } else {
    incr top
    $w selection set -only $anchor $top,$left
    $w selection set $anchor $top,$right
    $w see $top,$vleft
  }
}

# proc:  ::ss::shift_left $w
# title: Handle the Shift-Left key
#
# Extend the selection to the left.
#
proc shift_left {w} {
  if {[$w cget -state]=="disabled"} return
  set bbox [$w selrect bbox]
  if {$bbox==""} return
  foreach {top left bottom right} $bbox break
  set anchor [$w index active]
  if {[scan $anchor {%d,%d} arow acol]!=2} return
  scan [$w index topleft] {%d,%d} vtop vleft
  if {$left<$acol} {
    if {$left==1} return
    incr left -1
    $w selection set $anchor $top,$left
    $w selection set $anchor $bottom,$left
    $w see $vtop,$left
  } else {
    incr right -1
    $w selection set -only $anchor $top,$right
    $w selection set $anchor $bottom,$right
    $w see $vtop,$right
  }
}

# proc:  ::ss::shift_right $w
# title: Handle the Shift-Right key
#
# Extend the selection to the right.
#
proc shift_right {w} {
  if {[$w cget -state]=="disabled"} return
  set bbox [$w selrect bbox]
  if {$bbox==""} return
  foreach {top left bottom right} $bbox break
  set anchor [$w index active]
  set lastcol [$w cget -cols]
  if {[scan $anchor {%d,%d} arow acol]!=2} return
  scan [$w index topleft] {%d,%d} vtop vleft
  if {$right>$acol} {
    if {$right==$lastcol} return
    incr right
    $w selection set $anchor $top,$right
    $w selection set $anchor $bottom,$right
    $w see $vtop,$right
  } else {
    incr left
    $w selection set -only $anchor $top,$left
    $w selection set $anchor $bottom,$left
    $w see $vtop,$left
  }
}

# proc:  ::ss::control_left $w
# title: Handle the Control-Left key
#
# Change the insertion character to the leftmost position in the cell.
#
proc control_left {w} {
  if {[$w cget -state]=="disabled"} return
  $w icursor 0
}

# proc:  ::ss::control_right $w
# title: Handle the Control-Right key
#
# Change the insertion character to the rightmost position in the cell.
#
proc control_right {w} {
  if {[$w cget -state]=="disabled"} return
  set end [string length [$w curvalue]]
  $w icursor $end
}

# proc:  ::ss::move_cell $w $x $y
# title: Change the active cell
#
# Change the active cell the specified number of cells.
# The new active cell is changed to be $x cells away in the X direction
# and $y cells away in the Y direction.
#
proc move_cell {w x y} {
  set old_active [$w index active]
  if {[scan $old_active {%d,%d} orow ocol]!=2} return

  set last_row [expr {[$w cget -maxrows] - 1}]
  set newrow [expr {$orow + $y}]
  if {$newrow > $last_row} {
    set newrow $last_row
  } elseif {$newrow < 1} {
    set newrow 1
  }

  set ncols [$w cget -cols]
  set last_col [expr {[$w cget -cols] - 1}]
  set newcol [expr {$ocol + $x}]
  if {$newcol > $last_col} {
    set newcol $last_col
  } elseif {$newcol < 1} {
    set newcol 1
  }
  $w see $newrow,$newcol
  $w activate $newrow,$newcol
  $w selection set -only $newrow,$newcol
  $w trim
}

# proc:  ::ss::page_down $w
# title: Handle the Page Down key
#
# Handle page down.
# The currently active cell changes.
# This is Excel behavior.
# Note that this only happens when the user uses the keyboard;
# using the scrollbar doesn't change the active cell.
#
proc page_down {w} {
  if {[$w cget -state]=="disabled" && [$w cget -ifdisabled]=="all"} return
  set old_active [$w index active]

  $w yview scroll 1 pages
  if {$old_active==""} return
  if {[$w cget -state]=="disabled"} return

  set new_topleft [$w index topleft]
  scan $new_topleft {%d,%d} nrow ncol
  scan $old_active {%d,%d} arow acol
  $w activate $nrow,$acol
}

# proc:  ::ss::page_up $w
# title: Handle the Page Up key
#
proc page_up {w} {
  if {[$w cget -state]=="disabled" && [$w cget -ifdisabled]=="all"} return
  set old_active [$w index active]

  $w yview scroll -1 pages
  if {$old_active==""} return
  if {[$w cget -state]=="disabled"} return

  set new_topleft [$w index topleft]
  scan $new_topleft {%d,%d} nrow ncol
  scan $old_active {%d,%d} arow acol
  $w activate $nrow,$acol
}

# proc:  ::ss::control_home $w
# title: Handle the Control-Home key
#
proc control_home {w} {
  if {[$w cget -state]=="disabled" && [$w cget -ifdisabled]=="all"} return
  $w see 1,1
  if {[$w cget -state]=="disabled"} return
  $w selection set -only 1,1
  $w activate 1,1
}

# proc:  ::ss::control_shift_home $w
# title: Handle the Control-Shift-Home key
#
proc control_shift_home {w} {
  if {[$w cget -state]=="disabled" && [$w cget -ifdisabled]=="all"} return
  _extend_selection $w 1,1
}

# proc:  ::ss::control_end $w
# title: Handle the Control-End key
#
proc control_end {w} {
  if {[$w cget -state]=="disabled" && [$w cget -ifdisabled]=="all"} return
  set end [$w index end]
  if {[scan $end {%d,%d} row col]!=2} return
  $w see $row,1
  if {[$w cget -state]=="disabled"} return
  $w selection set -only $row,1
  $w activate $row,1
}

# proc:  ::ss::control_shift_end $w
# title: Handle the Control-Shift-End key
#
proc control_shift_end {w} {
  if {[$w cget -state]=="disabled" && [$w cget -ifdisabled]=="all"} return
  _extend_selection $w [$w index end]
}

# proc:  ::ss::b1_pressed $w $x $y
# title: Handle Button 1 press
#
# This proc is called when button 1 is pressed in the SS widget.
# [SRS4090]
# It activates the cell under $x,$y, and initiates a selection.
# [SRS3910] [SRS4040]
# If the button is pressed over a column or row header, then the entire column
# or row is selected.
# [SRS3930]
# If the button is pressed over the top-left unlabeled header cell,
# then the entire spreadsheet will be selected.
# [SRS3940] [SRS4070]
# There is no special handling for a double-click; it is treated as two
# single-clicks.
#
proc b1_pressed {w x y} {
  upvar #0 $w v
  if {[$w cget -state]=="disabled"} return

  # Order is important.  The selection must be set before
  # activating the cell.  Otherwise, application traces on the
  # widget variable will find the wrong selection set.

  _begin_select $w $x $y
  _begin_text_select $w $x $y
  $w activate @$x,$y
  focus $w
  set v(x) $x
  set v(y) $y
  set v(mouse_moved) 0
}

# proc:  ::ss::b1_pressed_can $w $x $y
# title: Handle Button 1 press in the canvas widget
#
# This proc is like ::ss::b1_pressed, but it handles the button press both
# inside the visible spreadsheet and also anywhere else in the canvas.
#
# It ignores presses within the canvas (they are handled by ::ss::b1_pressed).
# For presses outside the canvas, the selection is cleared.
#
proc b1_pressed_can {w x y} {
  set x [$w canvas canvasx $x]
  set y [$w canvas canvasy $y]
  if {[$w canvas find overlapping $x $y $x $y]==""} {
    $w selection clear all
  }
}

# proc:  ::ss::b1_motion $w $x $y
# title: Handle Button 1 Motion events
#
# [SRS3970] [SRS4030] [SRS4060] [SRS4120]
# This proc is called when the mouse is moved while button 1 is pressed.
# It extends the current selection, if the mouse has moved at least one pixel.
# [SRS4110]
# If the motion is confined to a single cell, then the intra-cell text selection
# is adjusted.
#
proc b1_motion {w x y} {
  upvar #0 $w v
  if {[$w cget -state]=="disabled"} return
  if {[set v(resize_line)]!=""} return
  if {[info exists v(sel_startx)]} return

  if {!$v(mouse_moved) && (abs($x - $v(x)) > 1 || abs($y - $v(y)) > 1)} {
    set v(mouse_moved) 1
  }
  if {!$v(mouse_moved)} return

  _move_selection $w [$w index @$x,$y]
  _move_text_selection $w $x $y
  set v(x) $x
  set v(y) $y
}

# proc:  ::ss::b1_released $w
# title: Handle Button 1 Release events
#
# This proc is called when button 1 is released.
# It cancels any auto-scan and commits the spreadsheet.
#
proc b1_released {w} {
  if {[$w cget -state]=="disabled"} return

  _cancel_auto_scan $w
  $w activate active
}

# proc:  ::ss::shift_b1_pressed $w $x $y
# title: Handle Shift-Button 1 press
#
# [SRS3920] [SRS4050]
# This proc is called when the user presses shift-button 1.
# It begins the extension of the current selection, if there is one;
# otherwise, it initiates a new selection.
#
proc shift_b1_pressed {w x y} {
  if {[$w cget -state]=="disabled"} return

  if {[$w selection includes anchor]} {
    _extend_selection $w [$w index @$x,$y]
    _extend_text_selection $w $x $y
  } else {
    b1_pressed $w $x $y
  }
}

# proc:  ::ss::control_b1_pressed $w $x $y
# title: Handle Control-Button 1 press
#
# This proc is called when the user presses control-button 1.
# It adds the cell under $x,$y to the selection set, and makes it active.
#
proc control_b1_pressed {w x y} {
  if {[$w cget -state]=="disabled"} return

  _add_selection $w [$w index @$x,$y]
}

# proc:  ::ss::b1_left $w $x $y
# title: Handle Leave events while button 1 is pressed
#
# This proc is called when the mouse with button 1 pressed leaves the
# spreadsheet.  It auto-scans the spreadsheet
#
proc b1_left {w x y} {
  upvar #0 $w v
  if {[$w cget -state]=="disabled"} return

  _cancel_auto_scan $w
  set v(x) $x
  set v(y) $y
  _poll_auto_scan $w
}

# proc:  ::ss::b1_entered $w
# title: Handle Enter events while button 1 is pressed
#
# This proc is called when the mouse with button 1 pressed enters the
# spreadsheet.  It cancels any previous auto-scan.
#
proc b1_entered {w} {
  _cancel_auto_scan $w
}

# proc:  ::ss::enter_vresize $w
# title: Handle Enter event into the vertical resize tag
#
proc enter_vresize {w} {
  upvar #0 $w v
  if {[$w cget -state]=="disabled" && [$w cget -ifdisabled]=="all"} return

  if {$v(resize_line)==""} {
    $w canvas config -cursor [$w cget -vresizecursor]
  }
}

# proc:  ::ss::leave_vresize $w
# title: Handle Leave event from the vertical resize tag
#
proc leave_vresize {w} {
  upvar #0 $w v
  if {[$w cget -state]=="disabled" && [$w cget -ifdisabled]=="all"} return

  if {$v(resize_line)==""} {
    $w canvas config -cursor {}
  }
}

# proc:  ::ss::b1_vresize $w $x $y
# title: Handle Button 1 press over a vertical resize bar.
#
# [SRS3950]
# This proc initiates a column resize.
#
proc b1_vresize {w x y} {
  upvar #0 $w v
  if {[$w cget -state]=="disabled" && [$w cget -ifdisabled]=="all"} return

  $w canvas config -cursor [$w cget -vresizecursor]

  # Get the column being resized.  Since the column tag on the current item
  # refers to the column to the right of the vertical resize bar, 
  # and we are actually resizing the column to the left of the resize bar, 
  # find the first unhidden column to the left of the resize bar.

  set current [$w canvas find withtag current]
  set tags [$w fast_canvas gettags $current]
  if {$tags==""} return
  set col {}
  foreach tag $tags {
    if {[scan $tag {RC%d} col]==1} break
  }
  if {$col==""} return
  scan [$w canvas coords $current] {%f %f %f %f} xcur y0 xcur y1

  # Get the vertical resize bar to the left of the current one.
  # We will constrain resizing so that the column size doesn't fall 
  # under the minimum.

  incr col -1
  while {[$w hide get $col] && $col>0} {incr col -1}
  if {$col==0} {
    set xleft 0
  } else {
    set left_id [lindex [$w fast_canvas find "RC$col"] 0]
    if {$left_id==""} return
    scan [$w canvas coords $left_id] {%f %f %f %f} xleft y0 xleft y1
  }


  # Create a "resize line" that will give the user feedback as to
  # where the column boundary will be

  set nrows [$w displayedrows]
  set last_row [expr {$nrows - 1}]
  set ybot [lindex [$w canvas cget -scrollregion] 3]

  set v(resize_line) [$w canvas create line $xcur 0 $xcur $ybot -dash 1]
  set v(resize_prev) [expr {$xcur+1}]
  set v(resize_min) [expr {$xleft + [$w minwidth $col] + 1}]
  set v(resize_col) $col
}

# proc:  ::ss::b1_motion_vresize $w $x $y
# title: Handle Motion events while pressing button 1 over vertical resize bar
#
# This proc continues a previously initiated column resize.
#
proc b1_motion_vresize {w x y} {
  upvar #0 $w v
  if {[$w cget -state]=="disabled" && [$w cget -ifdisabled]=="all"} return
  if {$v(resize_line)==""} return

  set cx [$w canvas canvasx $x]
  if {$cx < $v(resize_min)} {
    set cx $v(resize_min)
  }
  set xAmount [expr {int($cx - $v(resize_prev))}]
  if {$xAmount==0} return

  set v(resize_prev) $cx

  $w canvas move $v(resize_line) $xAmount 0

  # Stretch or shrink the title for the column being resized

  set col $v(resize_col)
  set cell_id [lindex [$w fast_canvas find "C0,$col"] 0]
  if {$cell_id!=""} {
    scan [$w canvas coords $cell_id] {%f %f %f %f} x0 y0 x1 y1
    $w canvas coords $cell_id $x0 $y0 $cx $y1
    set xmid [expr {($cx + $x0)/2}]

    set text_id [lindex [$w fast_canvas find "0,$col"] 0]
    if {$text_id!=""} {
      set text_width [$w canvas itemcget $text_id -width]
      incr text_width $xAmount
      $w canvas itemconfig $text_id -width $text_width
      scan [$w canvas coords $text_id] {%f %f} x0 y0
      $w canvas coords $text_id $xmid $y0
    }
  }

  # Move all of the title canvas items that are beneath the column
  # being resized. This mimics Excel behavior

  set ncols [$w cget -cols]
  for {incr col} {$col < $ncols} {incr col} {
    foreach tag "RC$col C0,$col 0,$col" {
      set cid [lindex [$w fast_canvas find $tag] 0]
      if {$cid==""} continue
      $w canvas move $cid $xAmount 0
    }
  }
}

# proc:  ::ss::b1_release_vresize $w
# title: Handle Release event for a vertical resize
#
# This proc commits a column resize.
#
proc b1_release_vresize {w} {
  upvar #0 $w v
  if {[$w cget -state]=="disabled" && [$w cget -ifdisabled]=="all"} return

  $w canvas config -cursor {}
  $w canvas delete $v(resize_line)

  set col $v(resize_col)
  if {$col!=""} {
    set cell_id [lindex [$w fast_canvas find "C0,$col"] 0]
    if {$cell_id!=""} {
      scan [$w canvas coords $cell_id] {%f %f %f %f} x0 y0 x1 y1
      set width [expr {int($x1 - $x0 + 1)}]
      $w width $col $width
    }
  }

  set v(resize_line) {}
  set v(resize_min) -1
  set v(resize_prev) -1
  set v(resize_col) {}
}

# proc:  ::ss::enter_hresize $w
# title: Handle Enter event into a horizontal resize tag
#
proc enter_hresize {w} {
  upvar #0 $w v
  if {[$w cget -state]=="disabled" && [$w cget -ifdisabled]=="all"} return

  if {$v(resize_line)==""} {
    $w canvas config -cursor [$w cget -hresizecursor]
  }
}

# proc:  ::ss::leave_hresize $w
# title: Handle Leave event from a horizontal resize tag
#
proc leave_hresize {w} {
  upvar #0 $w v
  if {[$w cget -state]=="disabled" && [$w cget -ifdisabled]=="all"} return

  if {$v(resize_line)==""} {
    $w canvas config -cursor {}
  }
}

# proc:  ::ss::b1_hresize $w $x $y
# title: Handle Button 1 press over a horizontal resize bar
#
# This proc initiates a row resize.
#
proc b1_hresize {w x y} {
  upvar #0 $w v
  if {[$w cget -state]=="disabled" && [$w cget -ifdisabled]=="all"} return

  $w canvas config -cursor [$w cget -hresizecursor]

  # Get the row being resized.  Since the row tag on the current item
  # refers to the row below the horizontal resize bar, and we are actually
  # resizing the row above the resize bar, subtract one from the row tag.

  set current [$w canvas find withtag current]
  set tags [$w fast_canvas gettags $current]
  if {$tags==""} return
  set row {}
  foreach tag $tags {
    if {[scan $tag {RR%d} row]==1} break
  }
  if {$row==""} return
  incr row -1
  scan [$w canvas coords $current] {%f %f %f %f} x0 ycur x1 ycur

  # Get the horizontal resize bar above the current one.  We will constrain
  # resizing so that the row size doesn't fall under the minimum.

  set above_id [lindex [$w fast_canvas find "RR$row"] 0]
  if {$above_id==""} return
  scan [$w canvas coords $above_id] {%f %f %f %f} x0 yabove x1 yabove

  set min_rowheight [$w cget -minrowheight]
  set text_minrowheight [$w minheight $row]
  if {$text_minrowheight > $min_rowheight} {
    set min_rowheight $text_minrowheight
  }

  # Create a "resize line" that will give the user feedback as to
  # where the row boundary will be

  set ncols [$w cget -cols]
  set last_col [expr {$ncols - 1}]
  while {[$w hide get $last_col] && $last_col>0} {
    incr last_col -1
  }
  if {$last_col==0} return
  set xright [lindex [$w fast_canvas bbox C0,$last_col] 2]

  set v(resize_line) [$w canvas create line 0 $ycur $xright $ycur -dash 1]
  set v(resize_prev) [expr {$ycur + 1}]
  set v(resize_min) [expr {$yabove + $min_rowheight + 1}]
  set v(resize_row) $row
}

# proc:  ::ss::b1_motion_hresize $w $x $y
# title: Handle Motion events for a horizontal resize
#
# This proc continues a previously initiated row resize.
#
proc b1_motion_hresize {w x y} {
  upvar #0 $w v
  if {[$w cget -state]=="disabled" && [$w cget -ifdisabled]=="all"} return
  if {$v(resize_line)==""} return

  set cy [$w canvas canvasy $y]
  if {$cy < $v(resize_min)} {
    set cy $v(resize_min)
  }
  set yAmount [expr {int($cy - $v(resize_prev))}]
  if {$yAmount==0} return

  set v(resize_prev) $cy

  $w canvas move $v(resize_line) 0 $yAmount

  # Stretch or shrink the title for the row being resized

  set row $v(resize_row)
  set cell_id [lindex [$w fast_canvas find "C$row,0"] 0]
  if {$cell_id!=""} {
    scan [$w canvas coords $cell_id] {%f %f %f %f} x0 y0 x1 y1
    $w canvas coords $cell_id $x0 $y0 $x1 $cy
    set ymid [expr {($cy + $y0)/2}]

    set text_id [lindex [$w fast_canvas find "$row,0"] 0]
    if {$text_id!=""} {
      scan [$w canvas coords $text_id] {%f %f} x0 y0
      $w canvas coords $text_id $x0 $ymid
    }
  }

  # Move all of the title canvas items that are beneath the row being resized.
  # This mimics Excel behavior

  set nrows [$w displayedrows]
  set done 0
  for {incr row} {$row < $nrows} {incr row} {
    foreach tag "RR$row C$row,0 $row,0" {
      set cid [lindex [$w fast_canvas find $tag] 0]
      if {$cid==""} {
        set done 1
        break
      }
      $w canvas move $cid 0 $yAmount
    }
    if {$done} break
  }
}

# proc:  ::ss::b1_release_hresize $w
# title: Handle Release event for a row resize
#
# This proc commits the row resize.
#
proc b1_release_hresize {w} {
  upvar #0 $w v
  if {[$w cget -state]=="disabled" && [$w cget -ifdisabled]=="all"} return

  $w canvas config -cursor {}
  $w canvas delete $v(resize_line)

  set row $v(resize_row)
  if {$row!=""} {
    set cell_id [lindex [$w fast_canvas find "C$row,0"] 0]
    if {$cell_id!=""} {
      scan [$w canvas coords $cell_id] {%f %f %f %f} x0 y0 x1 y1
      set height [expr {int($y1 - $y0)}]
      $w height $row $height
    }
  }

  set v(resize_line) {}
  set v(resize_min) -1
  set v(resize_prev) -1
  set v(resize_row) {}
}

# proc:  ::ss::enter_selrect $w
# title: Handle the mouse entering the selection rectangle
#
# When the mouse enters the current selection rectangle, 
# its cursor will be changed to give the user a hint that clicking on the
# selection rectangle will cause something to happen.
#
proc enter_selrect {w} {
  if {[$w cget -state]=="disabled"} return
  $w canvas config -cursor {}
}

# proc:  ::ss::leave_selrect $w
# title: Handle the mouse leaving the selection rectangle
#
# When the mouse leaves the current selection rectangle, 
# its cursor will change back to the normal spreadsheet cursor.
#
proc leave_selrect {w} {
  if {[$w cget -state]=="disabled"} return
  $w canvas config -cursor [$w cget -cursor]
}

# proc:  ::ss::b1_in_selrect $w $x $y
# Title: Handle button-1 press in the selected rectangle
#
# This proc is called when the user presses button 1 while over a
# selected border rectangle.  It will initiate a drag and drop of
# the selection.
#
proc b1_in_selrect {w x y} {
  upvar #0 $w v
  if {[$w cget -state]=="disabled"} return
  set rect_bbox [$w selrect bbox]
  if {$rect_bbox==""} return

  # Find the cell in the selection rectangle that's closest to $x,$y and
  # within the selection rectangle.
  # This is the "anchor cell" of the selection rectangle, used to determine
  # where to place the shadow selection rectangle in ::ss::b1_motion_selrect.

  set x [$w canvas canvasx $x]
  set y [$w canvas canvasy $y]
  set anchor [_get_nearest_cell $w $x $y]
  if {[scan $anchor {%d,%d} arow acol]!=2} return

  foreach {tlrow tlcol brrow brcol} $rect_bbox break
  if {$arow<$tlrow} {
    set arow $tlrow
  } elseif {$arow>$brrow} {
    set arow $brrow
  }
  if {$acol<$tlcol} {
    set acol $tlcol
  } elseif {$acol>$brcol} {
    set acol $brcol
  }

  set bbox [$w bbox $arow,$acol]
  if {$bbox==""} return

  # Save the origin of the anchor cell.
  # Also save the current time in clicks.  This is used for "hysteresis",
  # so the user won't accidentally delete data by doing random mouse clicks.

  set v(sel_anchor) $arow,$acol
  set v(sel_startx) [lindex $bbox 0]
  set v(sel_starty) [lindex $bbox 1]
  set v(sel_clicks) [clock clicks]
  set v(sel_newcell) {}
}

# proc: ::ss::b1_motion_selrect $w $x $y
# title: Handle button 1 movement while dragging the selection rectangle
#
# This proc will draw a "shadow selection rectangle" that shows where
# the drop will occur, given the current cursor position.
#
proc b1_motion_selrect {w x y} {
  upvar #0 $w v
  if {[$w cget -state]=="disabled"} return

  $w canvas delete shadow

  if {![info exists v(sel_startx)]} return
  set clockdiff [expr {[clock clicks] - $v(sel_clicks)}]
  if {$clockdiff<200000} return  ;# 2/10 of a second

  set rect_bbox [$w selrect bbox]
  if {$rect_bbox==""} return

  set v(sel_newcell) {}

  # Find the candidate anchor cell for the shadow rectangle.
  # It is determined by using the distance the cursor has moved from
  # the upperleft corner of the selection rectangle's anchor cell.

  scan $v(sel_anchor) {%d,%d} anchor_row anchor_col
  set abbox [$w bbox $v(sel_anchor)]
  if {$abbox==""} return

  set cx [$w canvas canvasx $x]
  set cy [$w canvas canvasy $y]
  set xdiff [expr {$cx - $v(sel_startx)}]
  set ydiff [expr {$cy - $v(sel_starty)}]
  set xnew [expr {[lindex $abbox 0] + $xdiff}]
  set ynew [expr {[lindex $abbox 1] + $ydiff}]
  set new_anchor [_get_nearest_cell $w $xnew $ynew]
  if {$new_anchor=="" || $new_anchor==$v(sel_anchor)} return

  # Using the candidate anchor cell, find the candidate topleft row and col.

  scan $new_anchor {%d,%d} new_anchor_row new_anchor_col
  set anchor_rdiff [expr {$new_anchor_row-$anchor_row}]
  set anchor_cdiff [expr {$new_anchor_col-$anchor_col}]

  foreach {row0 col0 row1 col1} $rect_bbox break
  set new_tlrow [expr {$row0+$anchor_rdiff}]
  set new_tlcol [expr {$col0+$anchor_cdiff}]

  # Adjust the placement of the shadow rectangle, 
  # if re-anchoring it at $new_anchor would force 
  # any of it off the spreadsheet.

  set newbot [expr {$new_tlrow + $row1 - $row0}]
  set maxrow [expr {[$w displayedrows]-1}]

  set topleft [$w index topleft]
  scan $topleft {%d,%d} tlrow tlcol

  if {$new_tlrow<1} {
    set rdiff [expr {1-$new_tlrow}]
    incr newbot $rdiff
    if {$newbot>$maxrow} return
    set new_tlrow 1
  } elseif {$newbot>$maxrow} {
    set rdiff [expr {$newbot-$maxrow}]
    incr new_tlrow -$rdiff
    if {$new_tlrow<1} return
    set newbot $maxrow
  }

  set maxcol [expr {[$w cget -cols]-1}]
  set newright [expr {$new_tlcol + $col1 - $col0}]
  if {$new_tlcol<1} {
    set cdiff [expr {1-$new_tlcol}]
    incr newright $cdiff
    if {$newright>$maxcol} return
    set new_tlcol 1
  } elseif {$newright>$maxcol} {
    set cdiff [expr {$newright-$maxcol}]
    incr new_tlcol -$cdiff
    if {$new_tlcol<1} return
    set newright $maxcol
  }

  # Draw the shadow selection rectangle, which indicates to the user the
  # place where the drop would occur.

  if {![$w selrect draw $new_tlrow,$new_tlcol $newbot,$newright \
      gray50 shadow]} return

  set v(sel_newcell) $new_tlrow,$new_tlcol
}

# proc: ::ss::b1_release_selrect $w
# title: Handle button 1 release while dragging the selection rectangle
#
# This proc will drop the selection onto the spreadsheet location currently
# indicated by the mouse cursor.  The selection will be copied into the
# target cells, the origin cell data will be deleted, and the selection
# and the active cursor will be reset.
#
proc b1_release_selrect {w} {
  upvar #0 $w v
  if {[$w cget -state]=="disabled"} return

  $w canvas delete shadow

  # Move the contents of the old selection to the destination rectangle,
  # starting with the cell contained in v(sel_newcell).
  # Then delete the old selection's contents and reset the selection.

  set var [$w cget -variable]
  set rect_bbox [$w selrect bbox]
  if {$var!="" && $rect_bbox!="" && [info exists v(sel_newcell)]} {
    foreach {oldrow0 oldcol0 oldrow1 oldcol1} $rect_bbox break
    set oldcell "$oldrow0,$oldcol0"
    set newcell $v(sel_newcell)
    if {$newcell!="" && $oldcell!=$newcell} {
      upvar #0 $var vvar
      for {set i 0; set row $oldrow0} {$row<=$oldrow1} {incr i; incr row} {
        for {set j 0; set col $oldcol0} {$col<=$oldcol1} {incr j; incr col} {
          if {[info exists vvar($row,$col)]} {
            set rect($i,$j) $vvar($row,$col)
            unset vvar($row,$col)
          }
        }
      }

      scan $newcell {%d,%d} newrow newcol
      set newbot [expr {$newrow + $oldrow1 - $oldrow0}]
      set newright [expr {$newcol + $oldcol1 - $oldcol0}]
      for {set i 0; set row $newrow} {$row<=$newbot} {incr i; incr row} {
        for {set j 0; set col $newcol} {$col<=$newright} {incr j; incr col} {
          if {[info exists rect($i,$j)]} {
            set vvar($row,$col) $rect($i,$j)
          } else {
            catch {unset vvar($row,$col)}
          }
        }
      }
      $w selection set -only $newcell $newbot,$newright
      $w activate $newcell
    }
  }

  foreach entry [array names v sel_*] {
    unset v($entry)
  }
}

# proc:  ::ss::enter_title $w
# title: Handle Enter event into a title tag
#
proc enter_title {w} {
  if {[$w cget -state]=="disabled"} return

  $w canvas config -cursor [$w cget -titlecursor]
}

# proc:  ::ss::leave_title $w
# title: Handle Leave event from a title tag
#
proc leave_title {w} {
  if {[$w cget -state]=="disabled"} return

  $w canvas config -cursor [$w cget -cursor]
}

#
# All procs below this point are called indirectly (not directly by a event)
#

# Initiate a selection with the cell under the window coordinates $x,$y
#
proc _begin_select {w x y} {
  upvar #0 $w v

  # First, initiate the table selection

  set cell [$w index @$x,$y]
  if {$cell==""} {
    $w selection clear all
    return
  }
  scan $cell {%d,%d} row col
  if {$row==0 && $col==0} {
    $w selection set all
    set anchor "1,1"
  } elseif {$row==0} {
    $w selection set -only $cell [$w index end row],$col
    set anchor "1,$col"
  } elseif {$col==0} {
    $w selection set -only $cell $row,[$w index end col]
    set anchor "$row,1"
  } else {
    $w selection set -only $cell
    set anchor $cell
  }
  $w selection anchor $anchor
  set v(orig_selection) $cell
  set v(prev_selection) $cell
}

# Move the selection to the $cell.
#
proc _move_selection {w cell} {
  upvar #0 $w v
  if {$cell==""} return

  set orig $v(orig_selection)
  if {$orig==""} {
    set v(orig_selection) $cell
    return
  }

  set prev $v(prev_selection)
  if {$prev==$cell} return
  if {$prev==""} {
    set prev $cell
  }
  scan $prev {%d,%d} prev_row prev_col

  scan $orig {%d,%d} orig_row orig_col
  scan $cell {%d,%d} row col
  if {$orig_row==0 && $orig_col==0} {  ;# do nothing if started at upperleft
  } elseif {$orig_row==0} {            ;# extend selection by entire columns
    if {$col!=$prev_col} {
      $w selection clear 0,$orig_col [$w index end row],$prev_col
      $w selection set 0,$orig_col [$w index end row],$col
    }
  } elseif {$orig_col==0} {                 ;# extend selection by entire rows
    if {$row!=$prev_row} {
      $w selection clear $orig_row,0 $prev_row,[$w index end col]
      $w selection set $orig_row,0 $row,[$w index end col]
    }
  } else {
    $w selection clear $orig $prev
    $w selection set $orig $cell
  }
  set v(prev_selection) $cell
}

# Extend the selection to the $cell.
# The behavior of this is somewhat different than that of ::ss::_move_selection;
# in particular, this can cause entire rows, columns, or the entire
# spreadsheet to be included into the selection set.
# (This is Excel's behavior.)
#
proc _extend_selection {w cell} {
  upvar #0 $w v
  if {$cell==""} return

  set anchor [$w index anchor]

  set prev $v(prev_selection)
  if {$prev==$cell} return
  if {$prev==""} {
    set prev $cell
  }
  scan $anchor {%d,%d} anchor_row anchor_col

  scan $cell {%d,%d} row col

  if {$row==0 && $col==0} {
    $w selection set all
  } elseif {$row==0} {
    $w selection set -only $cell [$w index end row],$anchor_col
  } elseif {$col==0} {
    $w selection set -only $cell $anchor_row,[$w index end col]
  } else {
    $w selection set -only $anchor $cell
  }

  set v(prev_selection) $cell
}

# Add $cell into the selection set, and make $cell the active cell.
# This is the Excel behavior for shift-click.
#
proc _add_selection {w cell} {
  upvar #0 $w v
  if {$cell==""} return

  scan $cell {%d,%d} row col
  if {$row==0 && $col==0} {
    $w selection set all
    set anchor "1,1"
  } elseif {$row==0} {
    $w selection set $cell [$w index end row],$col
    set anchor "1,$col"
  } elseif {$col==0} {
    $w selection set $cell $row,[$w index end col]
    set anchor "$row,1"
  } else {
    $w selection set $cell
    set anchor $cell
  }
  $w selection anchor $anchor
  $w activate $anchor
  set v(orig_selection) $anchor
  set v(prev_selection) $anchor
}

# Toggle a set of cells into, or out of, the current selection.
# $cell is the cell under the cursor.
#
# This is currently unused, as Excel does not seem to provide a toggling
# capability.
#
proc _toggle_selection {w cell} {
  upvar #0 $w v
  if {$cell==""} return
  $w selection anchor $cell
  set v(orig_selection) $cell

  set v(prev_selection) $cell
  scan $cell {%d,%d} row col
  if {$row==0 && $col==0} {
    set end end
  } elseif {$row==0} {
    set end [$w index end row],$col
  } elseif {$col==0} {
    set end $row,[$w index end col]
  } else {
    set end $cell
  }
  if {[$w selection includes $end]} {
    $w selection clear $cell $end
  } else {
    $w selection set $cell $end
  }
}

# Initiate a text selection (intra-cell) starting at window coordinate $x,$y
#
proc _begin_text_select {w x y} {
  upvar #0 $w v
  set cell [$w index @$x,$y]
  if {$cell==""} return
  $w canvas select clear

  scan $cell {%d,%d} row col
  if {$row==0 || $col==0} return

  set cx [$w canvas canvasx $x]
  set cy [$w canvas canvasy $y]
  if {[catch {set index [$w fast_canvas index $cell @$cx,$cy]} error]} {
    return
  }
  set id [lindex [$w fast_canvas find $cell] 0]
  if {$id!=""} {
    $w canvas select from $id $index
    set v(select_cell) $cell
  }
}

# Move the endpoint of the text selection (intra-cell) to the point under 
# the window coordinate $x,$y.
# NOTE: Some versions of tk don't do this right.  tk 8.0 works.
#
proc _move_text_selection {w x y} {
  upvar #0 $w v
  if {![info exists v(select_cell)]} return
  set id [lindex [$w fast_canvas find $v(select_cell)] 0]
  if {$id==""} return

  set cell_under_mouse [$w index @$x,$y]
  if {$cell_under_mouse!=$v(select_cell)} {
    $w canvas select clear
    set v(select_cell) {}
  } else {
    set cx [$w canvas canvasx $x]
    set cy [$w canvas canvasy $y]
    set index [$w canvas index $id @$cx,$cy]
    $w canvas select to $id $index
  }
}

# Extend the text (intra-cell) selection to the point in the current
# selection nearest the window coordinate $x,$y.
# NOTE: Some versions of tk don't do this right.  tk 8.0 works.
#
proc _extend_text_selection {w x y} {
  upvar #0 $w v
  if {![info exists v(select_cell)]} return
  set id [lindex [$w fast_canvas find $v(select_cell)] 0]
  if {$id==""} return

  set cell_under_mouse [$w index @$x,$y]
  if {$cell_under_mouse!=$v(select_cell)} {
    $w canvas select clear
    set v(select_cell) {}
  } else {
    set cx [$w canvas canvasx $x]
    set cy [$w canvas canvasy $y]
    set index [$w canvas index $id @$cx,$cy]
    $w canvas select adjust $id $index
  }
}

# Auto-scan the spreadsheet by scrolling it in the direction that the mouse
# currently is, relative to the spreadsheet.
# This reschedules itself, so the existence of $w has to be checked
#
proc _poll_auto_scan {w} {
  upvar #0 $w v
  if {![winfo exists $w]} return

  foreach {xscroll yscroll} [_auto_scan $w $v(x) $v(y)] break
  set v(auto_scan_timer) [after 50 ::ss::_poll_auto_scan $w]

  if {$v(orig_selection)!="" && ($xscroll || $yscroll)} {
    _move_selection $w [$w index @$v(x),$v(y)]
  }
}

# This proc scrolls the spreadsheet based on the cursor at window position
# $x,$y.  It returns a list with two values.  The first value is TRUE
# if x scrolling took place and FALSE otherwise.  The second is the
# equivalent boolean value for the y scroll.
#
proc _auto_scan {w x y} {
  set yscroll 0
  if {$y >= [winfo height $w]} {
    $w yview scroll 1 units
    set yscroll 1
  } elseif {$y < 0} {
    $w yview scroll -1 units
    set yscroll 1
  }
  set xscroll 0
  if {$x >= [winfo width $w]} {
    $w xview scroll 1 units
    set xscroll 1
  } elseif {$x < 0} {
    $w xview scroll -1 units
    set xscroll 1
  }
  return [list $xscroll $yscroll]
}

# Cancel an existing auto-scan operation
#
proc _cancel_auto_scan {w} {
  upvar #0 $w v
  if {![info exists v(auto_scan_timer)]} return
  after cancel $v(auto_scan_timer)
  unset v(auto_scan_timer)
}

# Given a point $x, $y in canvas coordinates,
# find the visible spreadsheet cell that is closest to that point.
#
proc _get_nearest_cell {w x y} {

  # Bound $x,$y to the visible spreadsheet.

  set ss_bbox [$w canvas bbox cell]
  set ss_x0 [lindex $ss_bbox 0]
  set ss_x1 [lindex $ss_bbox 2]
  if {$x<=$ss_x0} {
    set x [expr {$ss_x0+1}]
  } elseif {$x>=$ss_x1} {
    set x [expr {$ss_x1-1}]
  }
  set ss_y0 [lindex $ss_bbox 1]
  set ss_y1 [lindex $ss_bbox 3]
  if {$y<=$ss_y0} {
    set y [expr {$ss_y0+1}]
  } elseif {$y>=$ss_y1} {
    set y [expr {$ss_y1-1}]
  }

  # $x and $y are in canvas coordinates.  They must be converted
  # to window coordinates in order to use '$w index'.

  set dx [expr {int([$w canvas canvasx 0])}]
  set dy [expr {int([$w canvas canvasy 0])}]
  set wx [expr {$x-$dx}]
  set wy [expr {$y-$dy}]
  set cell [$w index @$wx,$wy]
  if {[scan $cell {%d,%d} row col]!=2} return
  if {$row<1} {return {}}
  if {$col<1} {
    set col 1
  }
  return $row,$col
}

} ;# end of the "ss" namespace
